-module(core_parse).
-define(THIS_MODULE, core_parse).
-export([parse/1, parse_and_scan/1, format_error/1]).

-export([abstract/1,abstract/2,normalise/1]).

-include("core_parse.hrl").

tok_val(T) -> element(3, T).
tok_line(T) -> element(2, T).

abstract(T, _N) -> abstract(T).

abstract(Term) -> core_lib:make_literal(Term).

normalise(Core) -> core_lib:literal_value(Core).

%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% The parser generator will insert appropriate declarations before this line.%

parse(Tokens) ->
    case catch yeccpars1(Tokens, false, 0, [], []) of
	error ->
	    Errorline =
		if Tokens == [] -> 0; true -> element(2, hd(Tokens)) end,
	    {error,
	     {Errorline, ?THIS_MODULE, "syntax error at or after this line."}};
	Other ->
	    Other
    end.

parse_and_scan({Mod, Fun, Args}) ->
    case apply(Mod, Fun, Args) of
	{eof, _} ->
	    {ok, eof};
	{error, Descriptor, _} ->
	    {error, Descriptor};
	{ok, Tokens, _} ->
	    yeccpars1(Tokens, {Mod, Fun, Args}, 0, [], [])
    end.

format_error(Message) ->
    case io_lib:deep_char_list(Message) of
	true ->
	    Message;
	_ ->
	    io_lib:write(Message)
    end.

% To be used in grammar files to throw an error message to the parser toplevel.
% Doesn't have to be exported!
return_error(Line, Message) ->
    throw({error, {Line, ?THIS_MODULE, Message}}).


% Don't change yeccpars1/6 too much, it is called recursively by yeccpars2/8!
yeccpars1([Token | Tokens], Tokenizer, State, States, Vstack) ->
    yeccpars2(State, element(1, Token), States, Vstack, Token, Tokens,
	      Tokenizer);
yeccpars1([], {M, F, A}, State, States, Vstack) ->
    case catch apply(M, F, A) of
        {eof, Endline} ->
            {error, {Endline, ?THIS_MODULE, "end_of_file"}};
        {error, Descriptor, _Endline} ->
            {error, Descriptor};
        {'EXIT', Reason} ->
            {error, {0, ?THIS_MODULE, Reason}};
        {ok, Tokens, _Endline} ->
	    case catch yeccpars1(Tokens, {M, F, A}, State, States, Vstack) of
		error ->
		    Errorline = element(2, hd(Tokens)),
		    {error, {Errorline, ?THIS_MODULE,
			     "syntax error at or after this line."}};
		Other ->
		    Other
	    end
    end;
yeccpars1([], false, State, States, Vstack) ->
    yeccpars2(State, '$end', States, Vstack, {'$end', 999999}, [], false).

% For internal use only.
yeccerror(Token) ->
    {error,
     {element(2, Token), ?THIS_MODULE,
      ["syntax error before: ", yecctoken2string(Token)]}}.

yecctoken2string({atom, _, A}) -> io_lib:write(A);
yecctoken2string({integer,_,N}) -> io_lib:write(N);
yecctoken2string({float,_,F}) -> io_lib:write(F);
yecctoken2string({char,_,C}) -> io_lib:write_char(C);
yecctoken2string({var,_,V}) -> io_lib:format('~s', [V]);
yecctoken2string({string,_,S}) -> io_lib:write_string(S);
yecctoken2string({reserved_symbol, _, A}) -> io_lib:format('~w', [A]);
yecctoken2string({_Cat, _, Val}) -> io_lib:format('~w', [Val]);

yecctoken2string({'dot', _}) -> io_lib:format('~w', ['.']);
yecctoken2string({'$end', _}) ->
    [];
yecctoken2string({Other, _}) when atom(Other) ->
    io_lib:format('~w', [Other]);
yecctoken2string(Other) ->
    io_lib:write(Other).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


yeccpars2(0, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, 'module', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 2, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(1, 'module', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 317, [1 | __Ss], [__T | __Stack]);
yeccpars2(1, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(2, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 4, [2 | __Ss], [__T | __Stack]);
yeccpars2(2, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(3, '$end', _, __Stack, _, _, _) ->
 {ok, hd(__Stack)};
yeccpars2(3, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(4, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 5, [4 | __Ss], [__T | __Stack]);
yeccpars2(4, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(5, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 310, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(6, 'attributes', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [6 | __Ss], [__T | __Stack]);
yeccpars2(6, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(7, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 280, [7 | __Ss], [__T | __Stack]);
yeccpars2(7, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(8, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [8 | __Ss], [__T | __Stack]);
yeccpars2(8, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [8 | __Ss], [__T | __Stack]);
yeccpars2(8, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(13, __Cat, [8 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(9, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [9 | __Ss], [__T | __Stack]);
yeccpars2(9, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(10, '=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [10 | __Ss], [__T | __Stack]);
yeccpars2(10, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(11, '/', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [11 | __Ss], [__T | __Stack]);
yeccpars2(11, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(12, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [12 | __Ss], [__T | __Stack]);
yeccpars2(12, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [12 | __Ss], [__T | __Stack]);
yeccpars2(12, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(17, __Cat, [12 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(13, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(module_defs, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(14, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(anno_function_name, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(15, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [15 | __Ss], [__T | __Stack]);
yeccpars2(15, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(16, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_module{name = #c_atom{val = tok_val(__2)}, exports = __3, attrs = __4, defs = __5},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(module_definition, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(17, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(function_definitions, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(18, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(19, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_fname{id = tok_val(__1), arity = tok_val(__3)},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(function_name, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(20, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(21, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [21 | __Ss], [__T | __Stack]);
yeccpars2(21, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(22, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_def{name = __1, val = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(function_definition, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(23, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(24, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(anno_fun, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(25, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(26, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [26 | __Ss], [__T | __Stack]);
yeccpars2(26, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(27, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 269, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(28, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 267, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(anno_variables, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(29, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [29 | __Ss], [__T | __Stack]);
yeccpars2(29, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(30, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_var{name = tok_val(__1)},
 yeccpars2(yeccgoto(variable, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(31, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(anno_variable, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(32, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(33, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(34, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, '>#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 252, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(35, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(36, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, '>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 245, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(37, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(38, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_fun{vars = __3, body = __6},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(fun_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(39, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(40, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(41, '/', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_atom{val = tok_val(__1)},
 yeccpars2(yeccgoto(atomic_literal, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(42, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(43, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(44, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(45, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(46, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(47, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(48, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(49, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(50, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_char{val = tok_val(__1)},
 yeccpars2(yeccgoto(atomic_literal, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(51, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(52, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(53, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(anno_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(54, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_float{val = tok_val(__1)},
 yeccpars2(yeccgoto(atomic_literal, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(55, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(56, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(57, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_int{val = tok_val(__1)},
 yeccpars2(yeccgoto(atomic_literal, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(58, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 213, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(59, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(60, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(210, __Cat, [60 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(61, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(62, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_nil{},
 yeccpars2(yeccgoto(atomic_literal, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(63, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 204, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(64, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(65, 'after', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 92, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 91, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(66, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(67, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(68, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(69, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_string{val = tok_val(__1)},
 yeccpars2(yeccgoto(atomic_literal, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(70, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(71, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(72, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(73, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(74, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(75, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(anno_expressions, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(76, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [76 | __Ss], [__T | __Stack]);
yeccpars2(76, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(77, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_tuple{es = []},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(tuple, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(78, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_tuple{es = __2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tuple, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(79, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(80, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(anno_expressions, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(81, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(82, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(83, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(84, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [84 | __Ss], [__T | __Stack]);
yeccpars2(84, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(85, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(86, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(87, __Cat, __Ss,  [__8,__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  if length(__5) == 2 -> #c_try{expr = __2, vars = __5, body = __8}; true -> return_error(tok_line(__4),"syntax error after (") end,
 __Nss = lists:nthtail(7, __Ss),
 yeccpars2(yeccgoto(try_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(88, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, '>#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 183, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 91, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(89, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 108, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 91, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(90, '>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 175, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 108, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 91, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(91, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 108, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 91, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(92, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(93, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 91, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(anno_clauses, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(94, 'after', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 92, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(95, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(clause_pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(96, '=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 155, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(anno_pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(97, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_atom{val = tok_val(__1)},
 yeccpars2(yeccgoto(atomic_literal, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(98, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(atomic_pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(99, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(other_pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(100, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(other_pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(101, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(anno_clause, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(102, 'when', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 157, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(103, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(other_pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(104, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(anno_pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(105, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   {T,A} = __2, #c_receive{clauses = [], timeout = T, action = A}
  end,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(receive_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(106, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(other_pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(107, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 108, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 91, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 111, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(108, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 91, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(109, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 113, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(anno_patterns, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(110, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 112, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(111, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_tuple{es = []},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(tuple_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(112, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_tuple{es = __2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tuple_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(113, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [113 | __Ss], [__T | __Stack]);
yeccpars2(113, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [113 | __Ss], [__T | __Stack]);
yeccpars2(113, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [113 | __Ss], [__T | __Stack]);
yeccpars2(113, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [113 | __Ss], [__T | __Stack]);
yeccpars2(113, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [113 | __Ss], [__T | __Stack]);
yeccpars2(113, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 108, [113 | __Ss], [__T | __Stack]);
yeccpars2(113, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [113 | __Ss], [__T | __Stack]);
yeccpars2(113, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [113 | __Ss], [__T | __Stack]);
yeccpars2(113, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 91, [113 | __Ss], [__T | __Stack]);
yeccpars2(113, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [113 | __Ss], [__T | __Stack]);
yeccpars2(113, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(114, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(anno_patterns, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(115, '=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 155, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(116, '-|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 152, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(117, '-|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 118, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(anno_variable, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(118, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 119, [118 | __Ss], [__T | __Stack]);
yeccpars2(118, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(119, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [119 | __Ss], [__T | __Stack]);
yeccpars2(119, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 135, [119 | __Ss], [__T | __Stack]);
yeccpars2(119, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [119 | __Ss], [__T | __Stack]);
yeccpars2(119, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [119 | __Ss], [__T | __Stack]);
yeccpars2(119, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 130, [119 | __Ss], [__T | __Stack]);
yeccpars2(119, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 131, [119 | __Ss], [__T | __Stack]);
yeccpars2(119, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [119 | __Ss], [__T | __Stack]);
yeccpars2(119, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [119 | __Ss], [__T | __Stack]);
yeccpars2(119, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(120, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 121, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(121, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  core_lib:set_anno(__2,__4),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(anno_variable, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(122, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [122 | __Ss], [__T | __Stack]);
yeccpars2(122, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 135, [122 | __Ss], [__T | __Stack]);
yeccpars2(122, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [122 | __Ss], [__T | __Stack]);
yeccpars2(122, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [122 | __Ss], [__T | __Stack]);
yeccpars2(122, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 130, [122 | __Ss], [__T | __Stack]);
yeccpars2(122, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 131, [122 | __Ss], [__T | __Stack]);
yeccpars2(122, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [122 | __Ss], [__T | __Stack]);
yeccpars2(122, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [122 | __Ss], [__T | __Stack]);
yeccpars2(122, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(123, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(annotation, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(124, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  tok_val(__1),
 yeccpars2(yeccgoto(atomic_constant, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(125, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(constant, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(126, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  tok_val(__1),
 yeccpars2(yeccgoto(atomic_constant, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(127, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(constant, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(128, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [128 | __Ss], [__T | __Stack]);
yeccpars2(128, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(constants, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(129, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 139, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(130, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  tok_val(__1),
 yeccpars2(yeccgoto(atomic_constant, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(131, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  tok_val(__1),
 yeccpars2(yeccgoto(atomic_constant, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(132, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(yeccgoto(atomic_constant, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(133, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  tok_val(__1),
 yeccpars2(yeccgoto(atomic_constant, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(134, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(constant, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(135, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [135 | __Ss], [__T | __Stack]);
yeccpars2(135, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 135, [135 | __Ss], [__T | __Stack]);
yeccpars2(135, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [135 | __Ss], [__T | __Stack]);
yeccpars2(135, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [135 | __Ss], [__T | __Stack]);
yeccpars2(135, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [135 | __Ss], [__T | __Stack]);
yeccpars2(135, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 130, [135 | __Ss], [__T | __Stack]);
yeccpars2(135, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 131, [135 | __Ss], [__T | __Stack]);
yeccpars2(135, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [135 | __Ss], [__T | __Stack]);
yeccpars2(135, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(136, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [136 | __Ss], [__T | __Stack]);
yeccpars2(136, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(137, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(tuple_constant, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(138, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  list_to_tuple(__2),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tuple_constant, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(139, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(annotation, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(140, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [140 | __Ss], [__T | __Stack]);
yeccpars2(140, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 135, [140 | __Ss], [__T | __Stack]);
yeccpars2(140, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [140 | __Ss], [__T | __Stack]);
yeccpars2(140, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [140 | __Ss], [__T | __Stack]);
yeccpars2(140, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 130, [140 | __Ss], [__T | __Stack]);
yeccpars2(140, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 131, [140 | __Ss], [__T | __Stack]);
yeccpars2(140, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [140 | __Ss], [__T | __Stack]);
yeccpars2(140, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(141, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(constants, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(142, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {nil,tok_line(__1)},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(nil, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(143, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [143 | __Ss], [__T | __Stack]);
yeccpars2(143, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 147, [143 | __Ss], [__T | __Stack]);
yeccpars2(143, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 145, [143 | __Ss], [__T | __Stack]);
yeccpars2(143, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(144, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 135, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 130, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 131, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(145, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(yeccgoto(tail_constant, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(146, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(cons_constant, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(147, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 135, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 130, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 131, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(148, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 149, [148 | __Ss], [__T | __Stack]);
yeccpars2(148, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(149, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tail_constant, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(150, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 147, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 145, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(151, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tail_constant, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(152, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 119, [152 | __Ss], [__T | __Stack]);
yeccpars2(152, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(153, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 154, [153 | __Ss], [__T | __Stack]);
yeccpars2(153, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(154, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  core_lib:set_anno(__2,__4),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(anno_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(155, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 108, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 91, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(156, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_alias{var = __1, pat = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(other_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(157, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(158, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 159, [158 | __Ss], [__T | __Stack]);
yeccpars2(158, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(159, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(160, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_clause{pats = __1, guard = __3, body = __5},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(clause, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(161, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   {T,A} = __3, #c_receive{clauses = __2, timeout = T, action = A}
  end,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(receive_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(162, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(anno_clauses, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(163, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 164, [163 | __Ss], [__T | __Stack]);
yeccpars2(163, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(164, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(165, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__2,__4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(timeout, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(166, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 167, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 170, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 168, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(167, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 108, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 91, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(168, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_nil{},
 yeccpars2(yeccgoto(tail_pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(169, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_cons{hd = __2, tl = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(cons_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(170, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [170 | __Ss], [__T | __Stack]);
yeccpars2(170, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [170 | __Ss], [__T | __Stack]);
yeccpars2(170, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [170 | __Ss], [__T | __Stack]);
yeccpars2(170, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [170 | __Ss], [__T | __Stack]);
yeccpars2(170, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [170 | __Ss], [__T | __Stack]);
yeccpars2(170, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 108, [170 | __Ss], [__T | __Stack]);
yeccpars2(170, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [170 | __Ss], [__T | __Stack]);
yeccpars2(170, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [170 | __Ss], [__T | __Stack]);
yeccpars2(170, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 91, [170 | __Ss], [__T | __Stack]);
yeccpars2(170, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [170 | __Ss], [__T | __Stack]);
yeccpars2(170, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(171, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 172, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(172, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tail_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(173, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 167, [173 | __Ss], [__T | __Stack]);
yeccpars2(173, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 170, [173 | __Ss], [__T | __Stack]);
yeccpars2(173, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 168, [173 | __Ss], [__T | __Stack]);
yeccpars2(173, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(174, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_cons{hd = __2, tl = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tail_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(175, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(clause_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(176, '>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 177, [176 | __Ss], [__T | __Stack]);
yeccpars2(176, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(177, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(clause_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(178, '-|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 180, [178 | __Ss], [__T | __Stack]);
yeccpars2(178, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(179, '-|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 152, [179 | __Ss], [__T | __Stack]);
yeccpars2(179, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(anno_pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(180, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 119, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(181, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 182, [181 | __Ss], [__T | __Stack]);
yeccpars2(181, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(182, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  core_lib:set_anno(__2,__4),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(anno_clause, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(183, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_binary{segs = []},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(binary_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(184, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(185, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 192, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(186, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 190, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(segment_patterns, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(187, '>#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 189, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(188, '=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars2(yeccgoto(anno_variable, hd(__Ss)), '=', __Ss, __Stack, __T, __Ts, __Tzr);
yeccpars2(188, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(189, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_binary{segs = __2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(binary_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(190, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 91, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(191, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(segment_patterns, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(192, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 91, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(193, '*', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 202, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__1,1},
 yeccpars2(yeccgoto(seg_size_pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(194, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 195, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(195, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 196, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(196, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  tok_val(__1),
 yeccpars2(yeccgoto(seg_type, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(197, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 198, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(199, __Cat, [197 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(198, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 200, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(199, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   {Size,Unit} = __3, #c_bin_seg{val = __1, size = Size, unit = Unit, type = __5, flags = __6}
  end,
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(segment_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(200, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 198, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(201, __Cat, [200 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(201, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [tok_val(__2)|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(seg_flags, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(202, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 203, [202 | __Ss], [__T | __Stack]);
yeccpars2(202, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(203, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__1,tok_val(__3)},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(seg_size_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(204, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 205, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(205, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 207, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(206, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   Name = #c_atom{val = tok_val(__2)}, #c_primop{name = Name, args = __3}
  end,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(primop_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(207, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(arg_list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(208, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 209, [208 | __Ss], [__T | __Stack]);
yeccpars2(208, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(209, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(arg_list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(210, 'in', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 211, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(211, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(212, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_letrec{defs = __2, body = __4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(letrec_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(213, '>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 220, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(214, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(let_vars, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(215, '=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 216, [215 | __Ss], [__T | __Stack]);
yeccpars2(215, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(216, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(217, 'in', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 218, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(218, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(219, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_let{vars = __2, arg = __4, body = __6},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(let_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(220, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(let_vars, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(221, '>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 222, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(222, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(let_vars, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(223, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(224, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_seq{arg = __2, body = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(sequence, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(225, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_catch{body = __2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(catch_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(226, 'of', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 227, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(227, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 91, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(228, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 229, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(229, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_case{arg = __2, clauses = __4},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(case_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(230, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 231, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(231, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(232, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 205, [232 | __Ss], [__T | __Stack]);
yeccpars2(232, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(233, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_call{module = __2, name = __4, args = __5},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(call_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(234, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 205, [234 | __Ss], [__T | __Stack]);
yeccpars2(234, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(235, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_apply{op = __2, args = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(application_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(236, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 237, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 240, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 238, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(237, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(238, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_nil{},
 yeccpars2(yeccgoto(tail, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(239, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_cons{hd = __2, tl = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(cons, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(240, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(241, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 242, [241 | __Ss], [__T | __Stack]);
yeccpars2(241, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(242, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tail, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(243, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 237, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 240, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 238, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(244, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_cons{hd = __2, tl = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tail, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(245, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_values{es = []},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(expression, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(246, '>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 247, [246 | __Ss], [__T | __Stack]);
yeccpars2(246, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(247, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_values{es = __2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expression, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(248, '-|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 249, [248 | __Ss], [__T | __Stack]);
yeccpars2(248, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(249, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 119, [249 | __Ss], [__T | __Stack]);
yeccpars2(249, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(250, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 251, [250 | __Ss], [__T | __Stack]);
yeccpars2(250, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(251, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  core_lib:set_anno(__2,__4),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(anno_expression, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(252, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_binary{segs = []},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(binary, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(253, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 265, [253 | __Ss], [__T | __Stack]);
yeccpars2(253, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(segments, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(254, '>#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 264, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(255, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 256, [255 | __Ss], [__T | __Stack]);
yeccpars2(255, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(256, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(257, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 261, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(258, '*', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 259, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__1,1},
 yeccpars2(yeccgoto(seg_size_unit, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(259, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 260, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(260, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__1,tok_val(__3)},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(seg_size_unit, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(261, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 196, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(262, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 198, [262 | __Ss], [__T | __Stack]);
yeccpars2(262, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(263, __Cat, [262 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(263, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   {Size,Unit} = __3, #c_bin_seg{val = __1, size = Size, unit = Unit, type = __5, flags = __6}
  end,
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(segment, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(264, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_binary{segs = __2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(binary, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(265, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(266, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(segments, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(267, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(268, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(anno_variables, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(269, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(270, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_fun{vars = [], body = __5},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(fun_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(271, '-|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 118, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(272, '-|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 273, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(273, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 119, [273 | __Ss], [__T | __Stack]);
yeccpars2(273, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(274, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 275, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(275, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  core_lib:set_anno(__2,__4),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(anno_fun, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(276, '-|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 277, [276 | __Ss], [__T | __Stack]);
yeccpars2(276, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(277, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 119, [277 | __Ss], [__T | __Stack]);
yeccpars2(277, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(278, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 279, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(279, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  core_lib:set_anno(__2,__4),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(anno_function_name, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(280, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 282, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 281, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(281, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(module_attribute, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(282, '=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 288, [282 | __Ss], [__T | __Stack]);
yeccpars2(282, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(283, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 286, [283 | __Ss], [__T | __Stack]);
yeccpars2(283, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(attribute_list, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(284, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 285, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(285, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __3,
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(module_attribute, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(286, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 282, [286 | __Ss], [__T | __Stack]);
yeccpars2(286, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(287, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(attribute_list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(288, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 289, [288 | __Ss], [__T | __Stack]);
yeccpars2(288, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 294, [288 | __Ss], [__T | __Stack]);
yeccpars2(288, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [288 | __Ss], [__T | __Stack]);
yeccpars2(288, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [288 | __Ss], [__T | __Stack]);
yeccpars2(288, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [288 | __Ss], [__T | __Stack]);
yeccpars2(288, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [288 | __Ss], [__T | __Stack]);
yeccpars2(288, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [288 | __Ss], [__T | __Stack]);
yeccpars2(288, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(289, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 289, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 294, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(290, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(literal, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(291, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(literal, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(292, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_def{name = #c_atom{val = tok_val(__1)}, val = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(attribute, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(293, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(literal, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(294, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 289, [294 | __Ss], [__T | __Stack]);
yeccpars2(294, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 294, [294 | __Ss], [__T | __Stack]);
yeccpars2(294, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 297, [294 | __Ss], [__T | __Stack]);
yeccpars2(294, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [294 | __Ss], [__T | __Stack]);
yeccpars2(294, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [294 | __Ss], [__T | __Stack]);
yeccpars2(294, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [294 | __Ss], [__T | __Stack]);
yeccpars2(294, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [294 | __Ss], [__T | __Stack]);
yeccpars2(294, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [294 | __Ss], [__T | __Stack]);
yeccpars2(294, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(295, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 299, [295 | __Ss], [__T | __Stack]);
yeccpars2(295, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(literals, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(296, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 298, [296 | __Ss], [__T | __Stack]);
yeccpars2(296, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(297, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_tuple{es = []},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(tuple_literal, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(298, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_tuple{es = __2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tuple_literal, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(299, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 289, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 294, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(300, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(literals, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(301, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 302, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 305, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 303, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(302, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 289, [302 | __Ss], [__T | __Stack]);
yeccpars2(302, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 294, [302 | __Ss], [__T | __Stack]);
yeccpars2(302, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [302 | __Ss], [__T | __Stack]);
yeccpars2(302, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [302 | __Ss], [__T | __Stack]);
yeccpars2(302, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [302 | __Ss], [__T | __Stack]);
yeccpars2(302, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [302 | __Ss], [__T | __Stack]);
yeccpars2(302, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [302 | __Ss], [__T | __Stack]);
yeccpars2(302, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(303, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_nil{},
 yeccpars2(yeccgoto(tail_literal, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(304, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_cons{hd = __2, tl = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(cons_literal, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(305, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 289, [305 | __Ss], [__T | __Stack]);
yeccpars2(305, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 294, [305 | __Ss], [__T | __Stack]);
yeccpars2(305, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [305 | __Ss], [__T | __Stack]);
yeccpars2(305, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [305 | __Ss], [__T | __Stack]);
yeccpars2(305, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [305 | __Ss], [__T | __Stack]);
yeccpars2(305, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [305 | __Ss], [__T | __Stack]);
yeccpars2(305, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [305 | __Ss], [__T | __Stack]);
yeccpars2(305, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(306, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 307, [306 | __Ss], [__T | __Stack]);
yeccpars2(306, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(307, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tail_literal, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(308, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 302, [308 | __Ss], [__T | __Stack]);
yeccpars2(308, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 305, [308 | __Ss], [__T | __Stack]);
yeccpars2(308, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 303, [308 | __Ss], [__T | __Stack]);
yeccpars2(308, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(309, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_cons{hd = __2, tl = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tail_literal, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(310, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(module_export, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(311, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 315, [311 | __Ss], [__T | __Stack]);
yeccpars2(311, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(exported_names, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(312, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 314, [312 | __Ss], [__T | __Stack]);
yeccpars2(312, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(313, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(exported_name, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(314, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(module_export, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(315, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [315 | __Ss], [__T | __Stack]);
yeccpars2(315, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(316, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(exported_names, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(317, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 318, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(318, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 5, [318 | __Ss], [__T | __Stack]);
yeccpars2(318, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(319, 'attributes', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [319 | __Ss], [__T | __Stack]);
yeccpars2(319, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(320, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [320 | __Ss], [__T | __Stack]);
yeccpars2(320, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [320 | __Ss], [__T | __Stack]);
yeccpars2(320, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(13, __Cat, [320 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(321, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 322, [321 | __Ss], [__T | __Stack]);
yeccpars2(321, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(322, '-|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 323, [322 | __Ss], [__T | __Stack]);
yeccpars2(322, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(323, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 119, [323 | __Ss], [__T | __Stack]);
yeccpars2(323, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(324, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 325, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(325, __Cat, __Ss,  [__10,__9,__8,__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_module{anno = __9, name = tok_val(__3), exports = __4, attrs = __5, defs = __6},
 __Nss = lists:nthtail(9, __Ss),
 yeccpars2(yeccgoto(module_definition, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(__Other, _, _, _, _, _, _) ->
 exit({parser, __Other, missing_state_in_action_table}).

yeccgoto(anno_clause, 65) ->
 93;
yeccgoto(anno_clause, 93) ->
 93;
yeccgoto(anno_clause, 227) ->
 93;
yeccgoto(anno_clauses, 65) ->
 94;
yeccgoto(anno_clauses, 93) ->
 162;
yeccgoto(anno_clauses, 227) ->
 228;
yeccgoto(anno_expression, 33) ->
 38;
yeccgoto(anno_expression, 36) ->
 75;
yeccgoto(anno_expression, 37) ->
 236;
yeccgoto(anno_expression, 40) ->
 234;
yeccgoto(anno_expression, 44) ->
 230;
yeccgoto(anno_expression, 46) ->
 226;
yeccgoto(anno_expression, 48) ->
 225;
yeccgoto(anno_expression, 52) ->
 223;
yeccgoto(anno_expression, 70) ->
 81;
yeccgoto(anno_expression, 74) ->
 75;
yeccgoto(anno_expression, 79) ->
 75;
yeccgoto(anno_expression, 86) ->
 87;
yeccgoto(anno_expression, 92) ->
 163;
yeccgoto(anno_expression, 157) ->
 158;
yeccgoto(anno_expression, 159) ->
 160;
yeccgoto(anno_expression, 164) ->
 165;
yeccgoto(anno_expression, 205) ->
 75;
yeccgoto(anno_expression, 211) ->
 212;
yeccgoto(anno_expression, 216) ->
 217;
yeccgoto(anno_expression, 218) ->
 219;
yeccgoto(anno_expression, 223) ->
 224;
yeccgoto(anno_expression, 231) ->
 232;
yeccgoto(anno_expression, 237) ->
 243;
yeccgoto(anno_expression, 240) ->
 241;
yeccgoto(anno_expression, 269) ->
 270;
yeccgoto(anno_expressions, 36) ->
 246;
yeccgoto(anno_expressions, 74) ->
 76;
yeccgoto(anno_expressions, 79) ->
 80;
yeccgoto(anno_expressions, 205) ->
 208;
yeccgoto(anno_fun, 20) ->
 22;
yeccgoto(anno_function_name, 8) ->
 10;
yeccgoto(anno_function_name, 12) ->
 10;
yeccgoto(anno_function_name, 60) ->
 10;
yeccgoto(anno_function_name, 320) ->
 10;
yeccgoto(anno_pattern, 65) ->
 95;
yeccgoto(anno_pattern, 89) ->
 95;
yeccgoto(anno_pattern, 90) ->
 109;
yeccgoto(anno_pattern, 91) ->
 166;
yeccgoto(anno_pattern, 93) ->
 95;
yeccgoto(anno_pattern, 107) ->
 109;
yeccgoto(anno_pattern, 113) ->
 109;
yeccgoto(anno_pattern, 155) ->
 156;
yeccgoto(anno_pattern, 167) ->
 173;
yeccgoto(anno_pattern, 170) ->
 171;
yeccgoto(anno_pattern, 227) ->
 95;
yeccgoto(anno_patterns, 90) ->
 176;
yeccgoto(anno_patterns, 107) ->
 110;
yeccgoto(anno_patterns, 113) ->
 114;
yeccgoto(anno_variable, 25) ->
 28;
yeccgoto(anno_variable, 58) ->
 214;
yeccgoto(anno_variable, 65) ->
 96;
yeccgoto(anno_variable, 83) ->
 28;
yeccgoto(anno_variable, 88) ->
 115;
yeccgoto(anno_variable, 89) ->
 96;
yeccgoto(anno_variable, 90) ->
 96;
yeccgoto(anno_variable, 91) ->
 96;
yeccgoto(anno_variable, 93) ->
 96;
yeccgoto(anno_variable, 107) ->
 96;
yeccgoto(anno_variable, 108) ->
 115;
yeccgoto(anno_variable, 113) ->
 96;
yeccgoto(anno_variable, 155) ->
 96;
yeccgoto(anno_variable, 167) ->
 96;
yeccgoto(anno_variable, 170) ->
 96;
yeccgoto(anno_variable, 190) ->
 115;
yeccgoto(anno_variable, 192) ->
 115;
yeccgoto(anno_variable, 213) ->
 28;
yeccgoto(anno_variable, 227) ->
 96;
yeccgoto(anno_variable, 267) ->
 28;
yeccgoto(anno_variables, 25) ->
 29;
yeccgoto(anno_variables, 83) ->
 84;
yeccgoto(anno_variables, 213) ->
 221;
yeccgoto(anno_variables, 267) ->
 268;
yeccgoto(annotation, 118) ->
 120;
yeccgoto(annotation, 152) ->
 153;
yeccgoto(annotation, 180) ->
 181;
yeccgoto(annotation, 249) ->
 250;
yeccgoto(annotation, 273) ->
 274;
yeccgoto(annotation, 277) ->
 278;
yeccgoto(annotation, 323) ->
 324;
yeccgoto(application_expr, 33) ->
 39;
yeccgoto(application_expr, 34) ->
 39;
yeccgoto(application_expr, 35) ->
 39;
yeccgoto(application_expr, 36) ->
 39;
yeccgoto(application_expr, 37) ->
 39;
yeccgoto(application_expr, 40) ->
 39;
yeccgoto(application_expr, 44) ->
 39;
yeccgoto(application_expr, 46) ->
 39;
yeccgoto(application_expr, 48) ->
 39;
yeccgoto(application_expr, 52) ->
 39;
yeccgoto(application_expr, 70) ->
 39;
yeccgoto(application_expr, 74) ->
 39;
yeccgoto(application_expr, 79) ->
 39;
yeccgoto(application_expr, 86) ->
 39;
yeccgoto(application_expr, 92) ->
 39;
yeccgoto(application_expr, 157) ->
 39;
yeccgoto(application_expr, 159) ->
 39;
yeccgoto(application_expr, 164) ->
 39;
yeccgoto(application_expr, 205) ->
 39;
yeccgoto(application_expr, 211) ->
 39;
yeccgoto(application_expr, 216) ->
 39;
yeccgoto(application_expr, 218) ->
 39;
yeccgoto(application_expr, 223) ->
 39;
yeccgoto(application_expr, 231) ->
 39;
yeccgoto(application_expr, 237) ->
 39;
yeccgoto(application_expr, 240) ->
 39;
yeccgoto(application_expr, 256) ->
 39;
yeccgoto(application_expr, 265) ->
 39;
yeccgoto(application_expr, 269) ->
 39;
yeccgoto(arg_list, 204) ->
 206;
yeccgoto(arg_list, 232) ->
 233;
yeccgoto(arg_list, 234) ->
 235;
yeccgoto(atomic_constant, 119) ->
 125;
yeccgoto(atomic_constant, 122) ->
 125;
yeccgoto(atomic_constant, 135) ->
 125;
yeccgoto(atomic_constant, 140) ->
 125;
yeccgoto(atomic_constant, 144) ->
 125;
yeccgoto(atomic_constant, 147) ->
 125;
yeccgoto(atomic_literal, 33) ->
 42;
yeccgoto(atomic_literal, 34) ->
 42;
yeccgoto(atomic_literal, 35) ->
 42;
yeccgoto(atomic_literal, 36) ->
 42;
yeccgoto(atomic_literal, 37) ->
 42;
yeccgoto(atomic_literal, 40) ->
 42;
yeccgoto(atomic_literal, 44) ->
 42;
yeccgoto(atomic_literal, 46) ->
 42;
yeccgoto(atomic_literal, 48) ->
 42;
yeccgoto(atomic_literal, 52) ->
 42;
yeccgoto(atomic_literal, 65) ->
 98;
yeccgoto(atomic_literal, 70) ->
 42;
yeccgoto(atomic_literal, 74) ->
 42;
yeccgoto(atomic_literal, 79) ->
 42;
yeccgoto(atomic_literal, 86) ->
 42;
yeccgoto(atomic_literal, 88) ->
 98;
yeccgoto(atomic_literal, 89) ->
 98;
yeccgoto(atomic_literal, 90) ->
 98;
yeccgoto(atomic_literal, 91) ->
 98;
yeccgoto(atomic_literal, 92) ->
 42;
yeccgoto(atomic_literal, 93) ->
 98;
yeccgoto(atomic_literal, 107) ->
 98;
yeccgoto(atomic_literal, 108) ->
 98;
yeccgoto(atomic_literal, 113) ->
 98;
yeccgoto(atomic_literal, 155) ->
 98;
yeccgoto(atomic_literal, 157) ->
 42;
yeccgoto(atomic_literal, 159) ->
 42;
yeccgoto(atomic_literal, 164) ->
 42;
yeccgoto(atomic_literal, 167) ->
 98;
yeccgoto(atomic_literal, 170) ->
 98;
yeccgoto(atomic_literal, 190) ->
 98;
yeccgoto(atomic_literal, 192) ->
 98;
yeccgoto(atomic_literal, 205) ->
 42;
yeccgoto(atomic_literal, 211) ->
 42;
yeccgoto(atomic_literal, 216) ->
 42;
yeccgoto(atomic_literal, 218) ->
 42;
yeccgoto(atomic_literal, 223) ->
 42;
yeccgoto(atomic_literal, 227) ->
 98;
yeccgoto(atomic_literal, 231) ->
 42;
yeccgoto(atomic_literal, 237) ->
 42;
yeccgoto(atomic_literal, 240) ->
 42;
yeccgoto(atomic_literal, 256) ->
 42;
yeccgoto(atomic_literal, 265) ->
 42;
yeccgoto(atomic_literal, 269) ->
 42;
yeccgoto(atomic_literal, 288) ->
 290;
yeccgoto(atomic_literal, 289) ->
 290;
yeccgoto(atomic_literal, 294) ->
 290;
yeccgoto(atomic_literal, 299) ->
 290;
yeccgoto(atomic_literal, 302) ->
 290;
yeccgoto(atomic_literal, 305) ->
 290;
yeccgoto(atomic_pattern, 65) ->
 99;
yeccgoto(atomic_pattern, 88) ->
 99;
yeccgoto(atomic_pattern, 89) ->
 99;
yeccgoto(atomic_pattern, 90) ->
 99;
yeccgoto(atomic_pattern, 91) ->
 99;
yeccgoto(atomic_pattern, 93) ->
 99;
yeccgoto(atomic_pattern, 107) ->
 99;
yeccgoto(atomic_pattern, 108) ->
 99;
yeccgoto(atomic_pattern, 113) ->
 99;
yeccgoto(atomic_pattern, 155) ->
 99;
yeccgoto(atomic_pattern, 167) ->
 99;
yeccgoto(atomic_pattern, 170) ->
 99;
yeccgoto(atomic_pattern, 190) ->
 99;
yeccgoto(atomic_pattern, 192) ->
 99;
yeccgoto(atomic_pattern, 227) ->
 99;
yeccgoto(attribute, 280) ->
 283;
yeccgoto(attribute, 286) ->
 283;
yeccgoto(attribute_list, 280) ->
 284;
yeccgoto(attribute_list, 286) ->
 287;
yeccgoto(binary, 33) ->
 43;
yeccgoto(binary, 34) ->
 43;
yeccgoto(binary, 35) ->
 43;
yeccgoto(binary, 36) ->
 43;
yeccgoto(binary, 37) ->
 43;
yeccgoto(binary, 40) ->
 43;
yeccgoto(binary, 44) ->
 43;
yeccgoto(binary, 46) ->
 43;
yeccgoto(binary, 48) ->
 43;
yeccgoto(binary, 52) ->
 43;
yeccgoto(binary, 70) ->
 43;
yeccgoto(binary, 74) ->
 43;
yeccgoto(binary, 79) ->
 43;
yeccgoto(binary, 86) ->
 43;
yeccgoto(binary, 92) ->
 43;
yeccgoto(binary, 157) ->
 43;
yeccgoto(binary, 159) ->
 43;
yeccgoto(binary, 164) ->
 43;
yeccgoto(binary, 205) ->
 43;
yeccgoto(binary, 211) ->
 43;
yeccgoto(binary, 216) ->
 43;
yeccgoto(binary, 218) ->
 43;
yeccgoto(binary, 223) ->
 43;
yeccgoto(binary, 231) ->
 43;
yeccgoto(binary, 237) ->
 43;
yeccgoto(binary, 240) ->
 43;
yeccgoto(binary, 256) ->
 43;
yeccgoto(binary, 265) ->
 43;
yeccgoto(binary, 269) ->
 43;
yeccgoto(binary_pattern, 65) ->
 100;
yeccgoto(binary_pattern, 88) ->
 100;
yeccgoto(binary_pattern, 89) ->
 100;
yeccgoto(binary_pattern, 90) ->
 100;
yeccgoto(binary_pattern, 91) ->
 100;
yeccgoto(binary_pattern, 93) ->
 100;
yeccgoto(binary_pattern, 107) ->
 100;
yeccgoto(binary_pattern, 108) ->
 100;
yeccgoto(binary_pattern, 113) ->
 100;
yeccgoto(binary_pattern, 155) ->
 100;
yeccgoto(binary_pattern, 167) ->
 100;
yeccgoto(binary_pattern, 170) ->
 100;
yeccgoto(binary_pattern, 190) ->
 100;
yeccgoto(binary_pattern, 192) ->
 100;
yeccgoto(binary_pattern, 227) ->
 100;
yeccgoto(call_expr, 33) ->
 45;
yeccgoto(call_expr, 34) ->
 45;
yeccgoto(call_expr, 35) ->
 45;
yeccgoto(call_expr, 36) ->
 45;
yeccgoto(call_expr, 37) ->
 45;
yeccgoto(call_expr, 40) ->
 45;
yeccgoto(call_expr, 44) ->
 45;
yeccgoto(call_expr, 46) ->
 45;
yeccgoto(call_expr, 48) ->
 45;
yeccgoto(call_expr, 52) ->
 45;
yeccgoto(call_expr, 70) ->
 45;
yeccgoto(call_expr, 74) ->
 45;
yeccgoto(call_expr, 79) ->
 45;
yeccgoto(call_expr, 86) ->
 45;
yeccgoto(call_expr, 92) ->
 45;
yeccgoto(call_expr, 157) ->
 45;
yeccgoto(call_expr, 159) ->
 45;
yeccgoto(call_expr, 164) ->
 45;
yeccgoto(call_expr, 205) ->
 45;
yeccgoto(call_expr, 211) ->
 45;
yeccgoto(call_expr, 216) ->
 45;
yeccgoto(call_expr, 218) ->
 45;
yeccgoto(call_expr, 223) ->
 45;
yeccgoto(call_expr, 231) ->
 45;
yeccgoto(call_expr, 237) ->
 45;
yeccgoto(call_expr, 240) ->
 45;
yeccgoto(call_expr, 256) ->
 45;
yeccgoto(call_expr, 265) ->
 45;
yeccgoto(call_expr, 269) ->
 45;
yeccgoto(case_expr, 33) ->
 47;
yeccgoto(case_expr, 34) ->
 47;
yeccgoto(case_expr, 35) ->
 47;
yeccgoto(case_expr, 36) ->
 47;
yeccgoto(case_expr, 37) ->
 47;
yeccgoto(case_expr, 40) ->
 47;
yeccgoto(case_expr, 44) ->
 47;
yeccgoto(case_expr, 46) ->
 47;
yeccgoto(case_expr, 48) ->
 47;
yeccgoto(case_expr, 52) ->
 47;
yeccgoto(case_expr, 70) ->
 47;
yeccgoto(case_expr, 74) ->
 47;
yeccgoto(case_expr, 79) ->
 47;
yeccgoto(case_expr, 86) ->
 47;
yeccgoto(case_expr, 92) ->
 47;
yeccgoto(case_expr, 157) ->
 47;
yeccgoto(case_expr, 159) ->
 47;
yeccgoto(case_expr, 164) ->
 47;
yeccgoto(case_expr, 205) ->
 47;
yeccgoto(case_expr, 211) ->
 47;
yeccgoto(case_expr, 216) ->
 47;
yeccgoto(case_expr, 218) ->
 47;
yeccgoto(case_expr, 223) ->
 47;
yeccgoto(case_expr, 231) ->
 47;
yeccgoto(case_expr, 237) ->
 47;
yeccgoto(case_expr, 240) ->
 47;
yeccgoto(case_expr, 256) ->
 47;
yeccgoto(case_expr, 265) ->
 47;
yeccgoto(case_expr, 269) ->
 47;
yeccgoto(catch_expr, 33) ->
 49;
yeccgoto(catch_expr, 34) ->
 49;
yeccgoto(catch_expr, 35) ->
 49;
yeccgoto(catch_expr, 36) ->
 49;
yeccgoto(catch_expr, 37) ->
 49;
yeccgoto(catch_expr, 40) ->
 49;
yeccgoto(catch_expr, 44) ->
 49;
yeccgoto(catch_expr, 46) ->
 49;
yeccgoto(catch_expr, 48) ->
 49;
yeccgoto(catch_expr, 52) ->
 49;
yeccgoto(catch_expr, 70) ->
 49;
yeccgoto(catch_expr, 74) ->
 49;
yeccgoto(catch_expr, 79) ->
 49;
yeccgoto(catch_expr, 86) ->
 49;
yeccgoto(catch_expr, 92) ->
 49;
yeccgoto(catch_expr, 157) ->
 49;
yeccgoto(catch_expr, 159) ->
 49;
yeccgoto(catch_expr, 164) ->
 49;
yeccgoto(catch_expr, 205) ->
 49;
yeccgoto(catch_expr, 211) ->
 49;
yeccgoto(catch_expr, 216) ->
 49;
yeccgoto(catch_expr, 218) ->
 49;
yeccgoto(catch_expr, 223) ->
 49;
yeccgoto(catch_expr, 231) ->
 49;
yeccgoto(catch_expr, 237) ->
 49;
yeccgoto(catch_expr, 240) ->
 49;
yeccgoto(catch_expr, 256) ->
 49;
yeccgoto(catch_expr, 265) ->
 49;
yeccgoto(catch_expr, 269) ->
 49;
yeccgoto(clause, 65) ->
 101;
yeccgoto(clause, 89) ->
 178;
yeccgoto(clause, 93) ->
 101;
yeccgoto(clause, 227) ->
 101;
yeccgoto(clause_pattern, 65) ->
 102;
yeccgoto(clause_pattern, 89) ->
 102;
yeccgoto(clause_pattern, 93) ->
 102;
yeccgoto(clause_pattern, 227) ->
 102;
yeccgoto(cons, 33) ->
 51;
yeccgoto(cons, 34) ->
 51;
yeccgoto(cons, 35) ->
 51;
yeccgoto(cons, 36) ->
 51;
yeccgoto(cons, 37) ->
 51;
yeccgoto(cons, 40) ->
 51;
yeccgoto(cons, 44) ->
 51;
yeccgoto(cons, 46) ->
 51;
yeccgoto(cons, 48) ->
 51;
yeccgoto(cons, 52) ->
 51;
yeccgoto(cons, 70) ->
 51;
yeccgoto(cons, 74) ->
 51;
yeccgoto(cons, 79) ->
 51;
yeccgoto(cons, 86) ->
 51;
yeccgoto(cons, 92) ->
 51;
yeccgoto(cons, 157) ->
 51;
yeccgoto(cons, 159) ->
 51;
yeccgoto(cons, 164) ->
 51;
yeccgoto(cons, 205) ->
 51;
yeccgoto(cons, 211) ->
 51;
yeccgoto(cons, 216) ->
 51;
yeccgoto(cons, 218) ->
 51;
yeccgoto(cons, 223) ->
 51;
yeccgoto(cons, 231) ->
 51;
yeccgoto(cons, 237) ->
 51;
yeccgoto(cons, 240) ->
 51;
yeccgoto(cons, 256) ->
 51;
yeccgoto(cons, 265) ->
 51;
yeccgoto(cons, 269) ->
 51;
yeccgoto(cons_constant, 119) ->
 127;
yeccgoto(cons_constant, 122) ->
 127;
yeccgoto(cons_constant, 135) ->
 127;
yeccgoto(cons_constant, 140) ->
 127;
yeccgoto(cons_constant, 144) ->
 127;
yeccgoto(cons_constant, 147) ->
 127;
yeccgoto(cons_literal, 288) ->
 291;
yeccgoto(cons_literal, 289) ->
 291;
yeccgoto(cons_literal, 294) ->
 291;
yeccgoto(cons_literal, 299) ->
 291;
yeccgoto(cons_literal, 302) ->
 291;
yeccgoto(cons_literal, 305) ->
 291;
yeccgoto(cons_pattern, 65) ->
 103;
yeccgoto(cons_pattern, 88) ->
 103;
yeccgoto(cons_pattern, 89) ->
 103;
yeccgoto(cons_pattern, 90) ->
 103;
yeccgoto(cons_pattern, 91) ->
 103;
yeccgoto(cons_pattern, 93) ->
 103;
yeccgoto(cons_pattern, 107) ->
 103;
yeccgoto(cons_pattern, 108) ->
 103;
yeccgoto(cons_pattern, 113) ->
 103;
yeccgoto(cons_pattern, 155) ->
 103;
yeccgoto(cons_pattern, 167) ->
 103;
yeccgoto(cons_pattern, 170) ->
 103;
yeccgoto(cons_pattern, 190) ->
 103;
yeccgoto(cons_pattern, 192) ->
 103;
yeccgoto(cons_pattern, 227) ->
 103;
yeccgoto(constant, 119) ->
 128;
yeccgoto(constant, 122) ->
 143;
yeccgoto(constant, 135) ->
 128;
yeccgoto(constant, 140) ->
 128;
yeccgoto(constant, 144) ->
 150;
yeccgoto(constant, 147) ->
 148;
yeccgoto(constants, 119) ->
 129;
yeccgoto(constants, 135) ->
 136;
yeccgoto(constants, 140) ->
 141;
yeccgoto(exported_name, 5) ->
 311;
yeccgoto(exported_name, 315) ->
 311;
yeccgoto(exported_names, 5) ->
 312;
yeccgoto(exported_names, 315) ->
 316;
yeccgoto(expression, 33) ->
 53;
yeccgoto(expression, 35) ->
 248;
yeccgoto(expression, 36) ->
 53;
yeccgoto(expression, 37) ->
 53;
yeccgoto(expression, 40) ->
 53;
yeccgoto(expression, 44) ->
 53;
yeccgoto(expression, 46) ->
 53;
yeccgoto(expression, 48) ->
 53;
yeccgoto(expression, 52) ->
 53;
yeccgoto(expression, 70) ->
 53;
yeccgoto(expression, 74) ->
 53;
yeccgoto(expression, 79) ->
 53;
yeccgoto(expression, 86) ->
 53;
yeccgoto(expression, 92) ->
 53;
yeccgoto(expression, 157) ->
 53;
yeccgoto(expression, 159) ->
 53;
yeccgoto(expression, 164) ->
 53;
yeccgoto(expression, 205) ->
 53;
yeccgoto(expression, 211) ->
 53;
yeccgoto(expression, 216) ->
 53;
yeccgoto(expression, 218) ->
 53;
yeccgoto(expression, 223) ->
 53;
yeccgoto(expression, 231) ->
 53;
yeccgoto(expression, 237) ->
 53;
yeccgoto(expression, 240) ->
 53;
yeccgoto(expression, 269) ->
 53;
yeccgoto(fun_expr, 20) ->
 24;
yeccgoto(fun_expr, 21) ->
 272;
yeccgoto(fun_expr, 33) ->
 55;
yeccgoto(fun_expr, 34) ->
 55;
yeccgoto(fun_expr, 35) ->
 55;
yeccgoto(fun_expr, 36) ->
 55;
yeccgoto(fun_expr, 37) ->
 55;
yeccgoto(fun_expr, 40) ->
 55;
yeccgoto(fun_expr, 44) ->
 55;
yeccgoto(fun_expr, 46) ->
 55;
yeccgoto(fun_expr, 48) ->
 55;
yeccgoto(fun_expr, 52) ->
 55;
yeccgoto(fun_expr, 70) ->
 55;
yeccgoto(fun_expr, 74) ->
 55;
yeccgoto(fun_expr, 79) ->
 55;
yeccgoto(fun_expr, 86) ->
 55;
yeccgoto(fun_expr, 92) ->
 55;
yeccgoto(fun_expr, 157) ->
 55;
yeccgoto(fun_expr, 159) ->
 55;
yeccgoto(fun_expr, 164) ->
 55;
yeccgoto(fun_expr, 205) ->
 55;
yeccgoto(fun_expr, 211) ->
 55;
yeccgoto(fun_expr, 216) ->
 55;
yeccgoto(fun_expr, 218) ->
 55;
yeccgoto(fun_expr, 223) ->
 55;
yeccgoto(fun_expr, 231) ->
 55;
yeccgoto(fun_expr, 237) ->
 55;
yeccgoto(fun_expr, 240) ->
 55;
yeccgoto(fun_expr, 256) ->
 55;
yeccgoto(fun_expr, 265) ->
 55;
yeccgoto(fun_expr, 269) ->
 55;
yeccgoto(function_definition, 8) ->
 12;
yeccgoto(function_definition, 12) ->
 12;
yeccgoto(function_definition, 60) ->
 12;
yeccgoto(function_definition, 320) ->
 12;
yeccgoto(function_definitions, 8) ->
 13;
yeccgoto(function_definitions, 12) ->
 17;
yeccgoto(function_definitions, 60) ->
 210;
yeccgoto(function_definitions, 320) ->
 13;
yeccgoto(function_name, 5) ->
 313;
yeccgoto(function_name, 8) ->
 14;
yeccgoto(function_name, 9) ->
 276;
yeccgoto(function_name, 12) ->
 14;
yeccgoto(function_name, 33) ->
 56;
yeccgoto(function_name, 34) ->
 56;
yeccgoto(function_name, 35) ->
 56;
yeccgoto(function_name, 36) ->
 56;
yeccgoto(function_name, 37) ->
 56;
yeccgoto(function_name, 40) ->
 56;
yeccgoto(function_name, 44) ->
 56;
yeccgoto(function_name, 46) ->
 56;
yeccgoto(function_name, 48) ->
 56;
yeccgoto(function_name, 52) ->
 56;
yeccgoto(function_name, 60) ->
 14;
yeccgoto(function_name, 70) ->
 56;
yeccgoto(function_name, 74) ->
 56;
yeccgoto(function_name, 79) ->
 56;
yeccgoto(function_name, 86) ->
 56;
yeccgoto(function_name, 92) ->
 56;
yeccgoto(function_name, 157) ->
 56;
yeccgoto(function_name, 159) ->
 56;
yeccgoto(function_name, 164) ->
 56;
yeccgoto(function_name, 205) ->
 56;
yeccgoto(function_name, 211) ->
 56;
yeccgoto(function_name, 216) ->
 56;
yeccgoto(function_name, 218) ->
 56;
yeccgoto(function_name, 223) ->
 56;
yeccgoto(function_name, 231) ->
 56;
yeccgoto(function_name, 237) ->
 56;
yeccgoto(function_name, 240) ->
 56;
yeccgoto(function_name, 256) ->
 56;
yeccgoto(function_name, 265) ->
 56;
yeccgoto(function_name, 269) ->
 56;
yeccgoto(function_name, 315) ->
 313;
yeccgoto(function_name, 320) ->
 14;
yeccgoto(let_expr, 33) ->
 59;
yeccgoto(let_expr, 34) ->
 59;
yeccgoto(let_expr, 35) ->
 59;
yeccgoto(let_expr, 36) ->
 59;
yeccgoto(let_expr, 37) ->
 59;
yeccgoto(let_expr, 40) ->
 59;
yeccgoto(let_expr, 44) ->
 59;
yeccgoto(let_expr, 46) ->
 59;
yeccgoto(let_expr, 48) ->
 59;
yeccgoto(let_expr, 52) ->
 59;
yeccgoto(let_expr, 70) ->
 59;
yeccgoto(let_expr, 74) ->
 59;
yeccgoto(let_expr, 79) ->
 59;
yeccgoto(let_expr, 86) ->
 59;
yeccgoto(let_expr, 92) ->
 59;
yeccgoto(let_expr, 157) ->
 59;
yeccgoto(let_expr, 159) ->
 59;
yeccgoto(let_expr, 164) ->
 59;
yeccgoto(let_expr, 205) ->
 59;
yeccgoto(let_expr, 211) ->
 59;
yeccgoto(let_expr, 216) ->
 59;
yeccgoto(let_expr, 218) ->
 59;
yeccgoto(let_expr, 223) ->
 59;
yeccgoto(let_expr, 231) ->
 59;
yeccgoto(let_expr, 237) ->
 59;
yeccgoto(let_expr, 240) ->
 59;
yeccgoto(let_expr, 256) ->
 59;
yeccgoto(let_expr, 265) ->
 59;
yeccgoto(let_expr, 269) ->
 59;
yeccgoto(let_vars, 58) ->
 215;
yeccgoto(letrec_expr, 33) ->
 61;
yeccgoto(letrec_expr, 34) ->
 61;
yeccgoto(letrec_expr, 35) ->
 61;
yeccgoto(letrec_expr, 36) ->
 61;
yeccgoto(letrec_expr, 37) ->
 61;
yeccgoto(letrec_expr, 40) ->
 61;
yeccgoto(letrec_expr, 44) ->
 61;
yeccgoto(letrec_expr, 46) ->
 61;
yeccgoto(letrec_expr, 48) ->
 61;
yeccgoto(letrec_expr, 52) ->
 61;
yeccgoto(letrec_expr, 70) ->
 61;
yeccgoto(letrec_expr, 74) ->
 61;
yeccgoto(letrec_expr, 79) ->
 61;
yeccgoto(letrec_expr, 86) ->
 61;
yeccgoto(letrec_expr, 92) ->
 61;
yeccgoto(letrec_expr, 157) ->
 61;
yeccgoto(letrec_expr, 159) ->
 61;
yeccgoto(letrec_expr, 164) ->
 61;
yeccgoto(letrec_expr, 205) ->
 61;
yeccgoto(letrec_expr, 211) ->
 61;
yeccgoto(letrec_expr, 216) ->
 61;
yeccgoto(letrec_expr, 218) ->
 61;
yeccgoto(letrec_expr, 223) ->
 61;
yeccgoto(letrec_expr, 231) ->
 61;
yeccgoto(letrec_expr, 237) ->
 61;
yeccgoto(letrec_expr, 240) ->
 61;
yeccgoto(letrec_expr, 256) ->
 61;
yeccgoto(letrec_expr, 265) ->
 61;
yeccgoto(letrec_expr, 269) ->
 61;
yeccgoto(literal, 288) ->
 292;
yeccgoto(literal, 289) ->
 301;
yeccgoto(literal, 294) ->
 295;
yeccgoto(literal, 299) ->
 295;
yeccgoto(literal, 302) ->
 308;
yeccgoto(literal, 305) ->
 306;
yeccgoto(literals, 294) ->
 296;
yeccgoto(literals, 299) ->
 300;
yeccgoto(module_attribute, 6) ->
 8;
yeccgoto(module_attribute, 319) ->
 320;
yeccgoto(module_definition, 0) ->
 3;
yeccgoto(module_defs, 8) ->
 15;
yeccgoto(module_defs, 320) ->
 321;
yeccgoto(module_export, 4) ->
 6;
yeccgoto(module_export, 318) ->
 319;
yeccgoto(nil, 33) ->
 62;
yeccgoto(nil, 34) ->
 62;
yeccgoto(nil, 35) ->
 62;
yeccgoto(nil, 36) ->
 62;
yeccgoto(nil, 37) ->
 62;
yeccgoto(nil, 40) ->
 62;
yeccgoto(nil, 44) ->
 62;
yeccgoto(nil, 46) ->
 62;
yeccgoto(nil, 48) ->
 62;
yeccgoto(nil, 52) ->
 62;
yeccgoto(nil, 65) ->
 62;
yeccgoto(nil, 70) ->
 62;
yeccgoto(nil, 74) ->
 62;
yeccgoto(nil, 79) ->
 62;
yeccgoto(nil, 86) ->
 62;
yeccgoto(nil, 88) ->
 62;
yeccgoto(nil, 89) ->
 62;
yeccgoto(nil, 90) ->
 62;
yeccgoto(nil, 91) ->
 62;
yeccgoto(nil, 92) ->
 62;
yeccgoto(nil, 93) ->
 62;
yeccgoto(nil, 107) ->
 62;
yeccgoto(nil, 108) ->
 62;
yeccgoto(nil, 113) ->
 62;
yeccgoto(nil, 119) ->
 132;
yeccgoto(nil, 122) ->
 132;
yeccgoto(nil, 135) ->
 132;
yeccgoto(nil, 140) ->
 132;
yeccgoto(nil, 144) ->
 132;
yeccgoto(nil, 147) ->
 132;
yeccgoto(nil, 155) ->
 62;
yeccgoto(nil, 157) ->
 62;
yeccgoto(nil, 159) ->
 62;
yeccgoto(nil, 164) ->
 62;
yeccgoto(nil, 167) ->
 62;
yeccgoto(nil, 170) ->
 62;
yeccgoto(nil, 190) ->
 62;
yeccgoto(nil, 192) ->
 62;
yeccgoto(nil, 205) ->
 62;
yeccgoto(nil, 211) ->
 62;
yeccgoto(nil, 216) ->
 62;
yeccgoto(nil, 218) ->
 62;
yeccgoto(nil, 223) ->
 62;
yeccgoto(nil, 227) ->
 62;
yeccgoto(nil, 231) ->
 62;
yeccgoto(nil, 237) ->
 62;
yeccgoto(nil, 240) ->
 62;
yeccgoto(nil, 256) ->
 62;
yeccgoto(nil, 265) ->
 62;
yeccgoto(nil, 269) ->
 62;
yeccgoto(nil, 288) ->
 62;
yeccgoto(nil, 289) ->
 62;
yeccgoto(nil, 294) ->
 62;
yeccgoto(nil, 299) ->
 62;
yeccgoto(nil, 302) ->
 62;
yeccgoto(nil, 305) ->
 62;
yeccgoto(other_pattern, 65) ->
 104;
yeccgoto(other_pattern, 88) ->
 184;
yeccgoto(other_pattern, 89) ->
 179;
yeccgoto(other_pattern, 90) ->
 104;
yeccgoto(other_pattern, 91) ->
 104;
yeccgoto(other_pattern, 93) ->
 104;
yeccgoto(other_pattern, 107) ->
 104;
yeccgoto(other_pattern, 108) ->
 116;
yeccgoto(other_pattern, 113) ->
 104;
yeccgoto(other_pattern, 155) ->
 104;
yeccgoto(other_pattern, 167) ->
 104;
yeccgoto(other_pattern, 170) ->
 104;
yeccgoto(other_pattern, 190) ->
 184;
yeccgoto(other_pattern, 192) ->
 184;
yeccgoto(other_pattern, 227) ->
 104;
yeccgoto(pattern, 88) ->
 185;
yeccgoto(pattern, 190) ->
 185;
yeccgoto(pattern, 192) ->
 193;
yeccgoto(primop_expr, 33) ->
 64;
yeccgoto(primop_expr, 34) ->
 64;
yeccgoto(primop_expr, 35) ->
 64;
yeccgoto(primop_expr, 36) ->
 64;
yeccgoto(primop_expr, 37) ->
 64;
yeccgoto(primop_expr, 40) ->
 64;
yeccgoto(primop_expr, 44) ->
 64;
yeccgoto(primop_expr, 46) ->
 64;
yeccgoto(primop_expr, 48) ->
 64;
yeccgoto(primop_expr, 52) ->
 64;
yeccgoto(primop_expr, 70) ->
 64;
yeccgoto(primop_expr, 74) ->
 64;
yeccgoto(primop_expr, 79) ->
 64;
yeccgoto(primop_expr, 86) ->
 64;
yeccgoto(primop_expr, 92) ->
 64;
yeccgoto(primop_expr, 157) ->
 64;
yeccgoto(primop_expr, 159) ->
 64;
yeccgoto(primop_expr, 164) ->
 64;
yeccgoto(primop_expr, 205) ->
 64;
yeccgoto(primop_expr, 211) ->
 64;
yeccgoto(primop_expr, 216) ->
 64;
yeccgoto(primop_expr, 218) ->
 64;
yeccgoto(primop_expr, 223) ->
 64;
yeccgoto(primop_expr, 231) ->
 64;
yeccgoto(primop_expr, 237) ->
 64;
yeccgoto(primop_expr, 240) ->
 64;
yeccgoto(primop_expr, 256) ->
 64;
yeccgoto(primop_expr, 265) ->
 64;
yeccgoto(primop_expr, 269) ->
 64;
yeccgoto(receive_expr, 33) ->
 66;
yeccgoto(receive_expr, 34) ->
 66;
yeccgoto(receive_expr, 35) ->
 66;
yeccgoto(receive_expr, 36) ->
 66;
yeccgoto(receive_expr, 37) ->
 66;
yeccgoto(receive_expr, 40) ->
 66;
yeccgoto(receive_expr, 44) ->
 66;
yeccgoto(receive_expr, 46) ->
 66;
yeccgoto(receive_expr, 48) ->
 66;
yeccgoto(receive_expr, 52) ->
 66;
yeccgoto(receive_expr, 70) ->
 66;
yeccgoto(receive_expr, 74) ->
 66;
yeccgoto(receive_expr, 79) ->
 66;
yeccgoto(receive_expr, 86) ->
 66;
yeccgoto(receive_expr, 92) ->
 66;
yeccgoto(receive_expr, 157) ->
 66;
yeccgoto(receive_expr, 159) ->
 66;
yeccgoto(receive_expr, 164) ->
 66;
yeccgoto(receive_expr, 205) ->
 66;
yeccgoto(receive_expr, 211) ->
 66;
yeccgoto(receive_expr, 216) ->
 66;
yeccgoto(receive_expr, 218) ->
 66;
yeccgoto(receive_expr, 223) ->
 66;
yeccgoto(receive_expr, 231) ->
 66;
yeccgoto(receive_expr, 237) ->
 66;
yeccgoto(receive_expr, 240) ->
 66;
yeccgoto(receive_expr, 256) ->
 66;
yeccgoto(receive_expr, 265) ->
 66;
yeccgoto(receive_expr, 269) ->
 66;
yeccgoto(seg_flags, 197) ->
 199;
yeccgoto(seg_flags, 200) ->
 201;
yeccgoto(seg_flags, 262) ->
 263;
yeccgoto(seg_size_pattern, 192) ->
 194;
yeccgoto(seg_size_unit, 256) ->
 257;
yeccgoto(seg_type, 195) ->
 197;
yeccgoto(seg_type, 261) ->
 262;
yeccgoto(segment, 34) ->
 253;
yeccgoto(segment, 265) ->
 253;
yeccgoto(segment_pattern, 88) ->
 186;
yeccgoto(segment_pattern, 190) ->
 186;
yeccgoto(segment_patterns, 88) ->
 187;
yeccgoto(segment_patterns, 190) ->
 191;
yeccgoto(segments, 34) ->
 254;
yeccgoto(segments, 265) ->
 266;
yeccgoto(sequence, 33) ->
 67;
yeccgoto(sequence, 34) ->
 67;
yeccgoto(sequence, 35) ->
 67;
yeccgoto(sequence, 36) ->
 67;
yeccgoto(sequence, 37) ->
 67;
yeccgoto(sequence, 40) ->
 67;
yeccgoto(sequence, 44) ->
 67;
yeccgoto(sequence, 46) ->
 67;
yeccgoto(sequence, 48) ->
 67;
yeccgoto(sequence, 52) ->
 67;
yeccgoto(sequence, 70) ->
 67;
yeccgoto(sequence, 74) ->
 67;
yeccgoto(sequence, 79) ->
 67;
yeccgoto(sequence, 86) ->
 67;
yeccgoto(sequence, 92) ->
 67;
yeccgoto(sequence, 157) ->
 67;
yeccgoto(sequence, 159) ->
 67;
yeccgoto(sequence, 164) ->
 67;
yeccgoto(sequence, 205) ->
 67;
yeccgoto(sequence, 211) ->
 67;
yeccgoto(sequence, 216) ->
 67;
yeccgoto(sequence, 218) ->
 67;
yeccgoto(sequence, 223) ->
 67;
yeccgoto(sequence, 231) ->
 67;
yeccgoto(sequence, 237) ->
 67;
yeccgoto(sequence, 240) ->
 67;
yeccgoto(sequence, 256) ->
 67;
yeccgoto(sequence, 265) ->
 67;
yeccgoto(sequence, 269) ->
 67;
yeccgoto(single_expression, 33) ->
 68;
yeccgoto(single_expression, 34) ->
 255;
yeccgoto(single_expression, 35) ->
 68;
yeccgoto(single_expression, 36) ->
 68;
yeccgoto(single_expression, 37) ->
 68;
yeccgoto(single_expression, 40) ->
 68;
yeccgoto(single_expression, 44) ->
 68;
yeccgoto(single_expression, 46) ->
 68;
yeccgoto(single_expression, 48) ->
 68;
yeccgoto(single_expression, 52) ->
 68;
yeccgoto(single_expression, 70) ->
 68;
yeccgoto(single_expression, 74) ->
 68;
yeccgoto(single_expression, 79) ->
 68;
yeccgoto(single_expression, 86) ->
 68;
yeccgoto(single_expression, 92) ->
 68;
yeccgoto(single_expression, 157) ->
 68;
yeccgoto(single_expression, 159) ->
 68;
yeccgoto(single_expression, 164) ->
 68;
yeccgoto(single_expression, 205) ->
 68;
yeccgoto(single_expression, 211) ->
 68;
yeccgoto(single_expression, 216) ->
 68;
yeccgoto(single_expression, 218) ->
 68;
yeccgoto(single_expression, 223) ->
 68;
yeccgoto(single_expression, 231) ->
 68;
yeccgoto(single_expression, 237) ->
 68;
yeccgoto(single_expression, 240) ->
 68;
yeccgoto(single_expression, 256) ->
 258;
yeccgoto(single_expression, 265) ->
 255;
yeccgoto(single_expression, 269) ->
 68;
yeccgoto(tail, 236) ->
 239;
yeccgoto(tail, 243) ->
 244;
yeccgoto(tail_constant, 143) ->
 146;
yeccgoto(tail_constant, 150) ->
 151;
yeccgoto(tail_literal, 301) ->
 304;
yeccgoto(tail_literal, 308) ->
 309;
yeccgoto(tail_pattern, 166) ->
 169;
yeccgoto(tail_pattern, 173) ->
 174;
yeccgoto(timeout, 65) ->
 105;
yeccgoto(timeout, 94) ->
 161;
yeccgoto(try_expr, 33) ->
 71;
yeccgoto(try_expr, 34) ->
 71;
yeccgoto(try_expr, 35) ->
 71;
yeccgoto(try_expr, 36) ->
 71;
yeccgoto(try_expr, 37) ->
 71;
yeccgoto(try_expr, 40) ->
 71;
yeccgoto(try_expr, 44) ->
 71;
yeccgoto(try_expr, 46) ->
 71;
yeccgoto(try_expr, 48) ->
 71;
yeccgoto(try_expr, 52) ->
 71;
yeccgoto(try_expr, 70) ->
 71;
yeccgoto(try_expr, 74) ->
 71;
yeccgoto(try_expr, 79) ->
 71;
yeccgoto(try_expr, 86) ->
 71;
yeccgoto(try_expr, 92) ->
 71;
yeccgoto(try_expr, 157) ->
 71;
yeccgoto(try_expr, 159) ->
 71;
yeccgoto(try_expr, 164) ->
 71;
yeccgoto(try_expr, 205) ->
 71;
yeccgoto(try_expr, 211) ->
 71;
yeccgoto(try_expr, 216) ->
 71;
yeccgoto(try_expr, 218) ->
 71;
yeccgoto(try_expr, 223) ->
 71;
yeccgoto(try_expr, 231) ->
 71;
yeccgoto(try_expr, 237) ->
 71;
yeccgoto(try_expr, 240) ->
 71;
yeccgoto(try_expr, 256) ->
 71;
yeccgoto(try_expr, 265) ->
 71;
yeccgoto(try_expr, 269) ->
 71;
yeccgoto(tuple, 33) ->
 72;
yeccgoto(tuple, 34) ->
 72;
yeccgoto(tuple, 35) ->
 72;
yeccgoto(tuple, 36) ->
 72;
yeccgoto(tuple, 37) ->
 72;
yeccgoto(tuple, 40) ->
 72;
yeccgoto(tuple, 44) ->
 72;
yeccgoto(tuple, 46) ->
 72;
yeccgoto(tuple, 48) ->
 72;
yeccgoto(tuple, 52) ->
 72;
yeccgoto(tuple, 70) ->
 72;
yeccgoto(tuple, 74) ->
 72;
yeccgoto(tuple, 79) ->
 72;
yeccgoto(tuple, 86) ->
 72;
yeccgoto(tuple, 92) ->
 72;
yeccgoto(tuple, 157) ->
 72;
yeccgoto(tuple, 159) ->
 72;
yeccgoto(tuple, 164) ->
 72;
yeccgoto(tuple, 205) ->
 72;
yeccgoto(tuple, 211) ->
 72;
yeccgoto(tuple, 216) ->
 72;
yeccgoto(tuple, 218) ->
 72;
yeccgoto(tuple, 223) ->
 72;
yeccgoto(tuple, 231) ->
 72;
yeccgoto(tuple, 237) ->
 72;
yeccgoto(tuple, 240) ->
 72;
yeccgoto(tuple, 256) ->
 72;
yeccgoto(tuple, 265) ->
 72;
yeccgoto(tuple, 269) ->
 72;
yeccgoto(tuple_constant, 119) ->
 134;
yeccgoto(tuple_constant, 122) ->
 134;
yeccgoto(tuple_constant, 135) ->
 134;
yeccgoto(tuple_constant, 140) ->
 134;
yeccgoto(tuple_constant, 144) ->
 134;
yeccgoto(tuple_constant, 147) ->
 134;
yeccgoto(tuple_literal, 288) ->
 293;
yeccgoto(tuple_literal, 289) ->
 293;
yeccgoto(tuple_literal, 294) ->
 293;
yeccgoto(tuple_literal, 299) ->
 293;
yeccgoto(tuple_literal, 302) ->
 293;
yeccgoto(tuple_literal, 305) ->
 293;
yeccgoto(tuple_pattern, 65) ->
 106;
yeccgoto(tuple_pattern, 88) ->
 106;
yeccgoto(tuple_pattern, 89) ->
 106;
yeccgoto(tuple_pattern, 90) ->
 106;
yeccgoto(tuple_pattern, 91) ->
 106;
yeccgoto(tuple_pattern, 93) ->
 106;
yeccgoto(tuple_pattern, 107) ->
 106;
yeccgoto(tuple_pattern, 108) ->
 106;
yeccgoto(tuple_pattern, 113) ->
 106;
yeccgoto(tuple_pattern, 155) ->
 106;
yeccgoto(tuple_pattern, 167) ->
 106;
yeccgoto(tuple_pattern, 170) ->
 106;
yeccgoto(tuple_pattern, 190) ->
 106;
yeccgoto(tuple_pattern, 192) ->
 106;
yeccgoto(tuple_pattern, 227) ->
 106;
yeccgoto(variable, 25) ->
 31;
yeccgoto(variable, 26) ->
 271;
yeccgoto(variable, 33) ->
 73;
yeccgoto(variable, 34) ->
 73;
yeccgoto(variable, 35) ->
 73;
yeccgoto(variable, 36) ->
 73;
yeccgoto(variable, 37) ->
 73;
yeccgoto(variable, 40) ->
 73;
yeccgoto(variable, 44) ->
 73;
yeccgoto(variable, 46) ->
 73;
yeccgoto(variable, 48) ->
 73;
yeccgoto(variable, 52) ->
 73;
yeccgoto(variable, 58) ->
 31;
yeccgoto(variable, 65) ->
 31;
yeccgoto(variable, 70) ->
 73;
yeccgoto(variable, 74) ->
 73;
yeccgoto(variable, 79) ->
 73;
yeccgoto(variable, 83) ->
 31;
yeccgoto(variable, 86) ->
 73;
yeccgoto(variable, 88) ->
 188;
yeccgoto(variable, 89) ->
 117;
yeccgoto(variable, 90) ->
 31;
yeccgoto(variable, 91) ->
 31;
yeccgoto(variable, 92) ->
 73;
yeccgoto(variable, 93) ->
 31;
yeccgoto(variable, 107) ->
 31;
yeccgoto(variable, 108) ->
 117;
yeccgoto(variable, 113) ->
 31;
yeccgoto(variable, 155) ->
 31;
yeccgoto(variable, 157) ->
 73;
yeccgoto(variable, 159) ->
 73;
yeccgoto(variable, 164) ->
 73;
yeccgoto(variable, 167) ->
 31;
yeccgoto(variable, 170) ->
 31;
yeccgoto(variable, 190) ->
 188;
yeccgoto(variable, 192) ->
 188;
yeccgoto(variable, 205) ->
 73;
yeccgoto(variable, 211) ->
 73;
yeccgoto(variable, 213) ->
 31;
yeccgoto(variable, 216) ->
 73;
yeccgoto(variable, 218) ->
 73;
yeccgoto(variable, 223) ->
 73;
yeccgoto(variable, 227) ->
 31;
yeccgoto(variable, 231) ->
 73;
yeccgoto(variable, 237) ->
 73;
yeccgoto(variable, 240) ->
 73;
yeccgoto(variable, 256) ->
 73;
yeccgoto(variable, 265) ->
 73;
yeccgoto(variable, 267) ->
 31;
yeccgoto(variable, 269) ->
 73;
yeccgoto(__Symbol, __State) ->
 exit({__Symbol, __State, missing_in_goto_table}).


